import bpy


def create_anim(ob, anim_name):
	action = bpy.data.actions.new(name=anim_name)
	action.use_fake_user = True
	ob.animation_data_create()
	ob.animation_data.action = action
	return action


class Animation:

	def __init__(self):
		self.fps = 30
		# store the actions per run here
		# we need to be able to map their names to blender actions
		# to prevent overwriting existing animations from older imports
		# and still be able to access existing actions from this run
		self.actions = {}

	@staticmethod
	def get_b_interp_from_n_interp(n_ipol):
		if n_ipol in (NifFormat.KeyType.LINEAR_KEY, NifFormat.KeyType.XYZ_ROTATION_KEY):
			return "LINEAR"
		elif n_ipol == NifFormat.KeyType.QUADRATIC_KEY:
			return "BEZIER"
		elif n_ipol == 0:
			# guessing, not documented in nif.xml
			return "CONSTANT"
		# NifLog.warn(f"Unsupported interpolation mode ({n_ipol}) in nif, using quadratic/bezier.")
		return "BEZIER"

	def create_action(self, b_obj, action_name):
		""" Create or retrieve action and set it as active on the object. """
		if action_name in self.actions:
			b_action = self.actions[action_name]
		else:
			b_action = bpy.data.actions.new(action_name)
			b_action.use_fake_user = True
			self.actions[action_name] = b_action
		# could probably skip this test and create always
		if not b_obj.animation_data:
			b_obj.animation_data_create()
		# set as active action on object
		b_obj.animation_data.action = b_action
		return b_action

	def create_fcurves(self, action, dtype, drange, flags=None, bonename=None, keyname=None):
		""" Create fcurves in action for desired conditions. """
		# armature pose bone animation
		if bonename:
			fcurves = [
				action.fcurves.new(data_path=f'pose.bones["{bonename}"].{dtype}', index=i, action_group=bonename)
				for i in drange]
		# shapekey pose bone animation
		elif keyname:
			fcurves = [
				action.fcurves.new(data_path=f'key_blocks["{keyname}"].{dtype}', index=0,)
			]
		else:
			# Object animation (non-skeletal) is lumped into the "LocRotScale" action_group
			if dtype in ("rotation_euler", "rotation_quaternion", "location", "scale"):
				action_group = "LocRotScale"
			# Non-transformaing animations (eg. visibility or material anims) use no action groups
			else:
				action_group = ""
			fcurves = [action.fcurves.new(data_path=dtype, index=i, action_group=action_group) for i in drange]
		if flags:
			self.set_extrapolation(self.get_extend_from_flags(flags), fcurves)
		return fcurves

	@staticmethod
	def get_extend_from_flags(flags):
		if flags & 6 == 4:  # 0b100
			return "CONSTANT"
		elif flags & 6 == 0:  # 0b000
			return "CYCLIC"

		NifLog.warn("Unsupported cycle mode in nif, using clamped.")
		return "CONSTANT"

	@staticmethod
	def get_extend_from_cycle_type(cycle_type):
		return ("CYCLIC", "REVERSE", "CONSTANT")[cycle_type]

	@staticmethod
	def set_extrapolation(extend_type, fcurves):
		if extend_type == "CONSTANT":
			for fcurve in fcurves:
				fcurve.extrapolation = 'CONSTANT'
		elif extend_type == "CYCLIC":
			for fcurve in fcurves:
				fcurve.modifiers.new('CYCLES')
		# don't support reverse for now, not sure if it is even possible in blender
		else:
			NifLog.warn("Unsupported extrapolation mode, using clamped.")
			for fcurve in fcurves:
				fcurve.extrapolation = 'CONSTANT'

	def add_key(self, fcurves, frame, key, interp):
		"""
		Add a key (len=n) to a set of fcurves (len=n) at the given frame. Set the key's interpolation to interp.
		"""
		# frame = round(t * self.fps)
		for fcurve, k in zip(fcurves, key):
			fcurve.keyframe_points.insert(frame, k)#.interpolation = interp

	# import animation groups
	def import_text_keys(self, n_block, b_action):
		"""Gets and imports a NiTextKeyExtraData"""
		if isinstance(n_block, NifFormat.NiControllerSequence):
			txk = n_block.text_keys
		else:
			txk = n_block.find(block_type=NifFormat.NiTextKeyExtraData)
		self.import_text_key_extra_data(txk, b_action)

	def import_text_key_extra_data(self, txk, b_action):
		"""Stores the text keys as pose markers in a blender action."""
		if txk and b_action:
			for key in txk.text_keys:
				newkey = key.value.decode().replace('\r\n', '/').rstrip('/')
				frame = round(key.time * self.fps)
				marker = b_action.pose_markers.new(newkey)
				marker.frame = frame

	def set_frames_per_second(self, roots):
		"""Scan all blocks and set a reasonable number for fps to this class and the scene."""
		# find all key times
		key_times = []
		for root in roots:
			for kfd in root.tree(block_type=NifFormat.NiKeyframeData):
				key_times.extend(key.time for key in kfd.translations.keys)
				key_times.extend(key.time for key in kfd.scales.keys)
				key_times.extend(key.time for key in kfd.quaternion_keys)
				key_times.extend(key.time for key in kfd.xyz_rotations[0].keys)
				key_times.extend(key.time for key in kfd.xyz_rotations[1].keys)
				key_times.extend(key.time for key in kfd.xyz_rotations[2].keys)

			for kfi in root.tree(block_type=NifFormat.NiBSplineInterpolator):
				if not kfi.basis_data:
					# skip bsplines without basis data (eg bowidle.kf in Oblivion)
					continue
				key_times.extend(
					point * (kfi.stop_time - kfi.start_time)
					/ (kfi.basis_data.num_control_points - 2)
					for point in range(kfi.basis_data.num_control_points - 2))

			for uv_data in root.tree(block_type=NifFormat.NiUVData):
				for uv_group in uv_data.uv_groups:
					key_times.extend(key.time for key in uv_group.keys)

		# not animated, return a reasonable default
		if not key_times:
			return

		# calculate fps
		key_times = sorted(set(key_times))
		fps = self.fps
		lowest_diff = sum(abs(int(time * fps + 0.5) - (time * fps)) for time in key_times)

		# for test_fps in range(1,120): #disabled, used for testing
		for test_fps in [20, 24, 25, 30, 35]:
			diff = sum(abs(int(time * test_fps + 0.5) - (time * test_fps)) for time in key_times)
			if diff < lowest_diff:
				lowest_diff = diff
				fps = test_fps
		NifLog.info(f"Animation estimated at {fps} frames per second.")
		self.fps = fps
		bpy.context.scene.render.fps = fps
		bpy.context.scene.frame_set(0)